#!/usr/bin/env python

"""
This script needs work!  Or will we never use it again?

There should be a --find option, or something.
"""

from __future__ import print_function

from dark.blast import BlastRecords
from dark.graphics import evalueGraph

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        print('Usage: %s BLAST-hitfile...' % sys.argv[0], file=sys.stderr)
    else:
        blastRecords = BlastRecords(sys.argv[1:])
        evalueGraph(blastRecords.records(), 10, 10,
                    find=lambda title: title.find('HKU4') > -1 and
                    title.find('complete genome') > -1, titles=False)



from collections import defaultdict, Counter

from dark.reads import Reads
from dark.utils import median
from dark.filter import ReadSetFilter
from dark.intervals import ReadIntervals


def titleCounts(readsAlignments):
    """
    Count the number of times each title in a readsAlignments instance is
    matched. This is useful for rapidly discovering what titles were matched
    and with what frequency.

    @param readsAlignments: A L{dark.alignments.ReadsAlignments} instance.
    @return: A C{dict} whose keys are titles and whose values are the integer
        counts of the number of reads that matched that title.
    """

    titles = defaultdict(int)
    for readAlignments in readsAlignments:
        for alignment in readAlignments:
            titles[alignment.subjectTitle] += 1
    return titles


class TitleAlignment(object):
    """
    Hold information about a read's HSPs for a title alignment.

    @param read: The C{Read} that aligned.
    @param hsps: A C{list} of L{dark.hsp.HSP} (or subclass) instance.
    """

    def __init__(self, read, hsps):
        self.read = read
        self.hsps = hsps


class TitleAlignments(list):
    """
    Holds information about a list of alignments against a sequence.

    @param subjectTitle: The C{str} title of the sequence the read matched
        against.
    @param subjectLength: The C{int} length of the sequence the read matched
        against.
    """

    def __init__(self, subjectTitle, subjectLength):
        # TODO: Do we need the title in here?
        self.subjectTitle = subjectTitle
        self.subjectLength = subjectLength

    def addAlignment(self, alignment):
        """
        Add an alignment to the list of alignments that matched this title.

        @param alignment: A L{TitleAlignment} instance.
        """
        self.append(alignment)

    def reads(self):
        """
        Find the set of reads matching this title.

        @return: An instance of C{dark.reads.Reads}.
        """
        reads = Reads()
        for alignment in self:
            reads.add(alignment.read)
        return reads

    def readCount(self):
        """
        Find out how many reads aligned to this title.

        @return: The C{int} number of reads that aligned to this title.
        """
        return len(self)

    def hspCount(self):
        """
        How many HSPs were there in total for all the alignments to a title.

        @return: The C{int} number of HSPs for the alignments to this title.
        """
        return sum(len(alignment.hsps) for alignment in self)

    def readIds(self):
        """
        Find the set of read ids that matched the title.

        @return: A C{set} of read ids that aligned to this title.
        """
        return set(alignment.read.id for alignment in self)

    def hsps(self):
        """
        Get all HSPs for the alignments to a title.

        @return: A generator yielding L{dark.hsp.HSP} instances.
        """
        return (hsp for alignment in self for hsp in alignment.hsps)

    def bestHsp(self):
        """
        Find the HSP with the best score.

        @raise ValueError: If there are no HSPs.
        @return: The C{dark.hsp.HSP} instance (or a subclass) with the best
        score.
        """
        return max(hsp for hsp in self.hsps())

    def worstHsp(self):
        """
        Find the HSP with the worst score.

        @raise ValueError: If there are no HSPs.
        @return: The C{dark.hsp.HSP} instance (or a subclass) with the worst
        score.
        """
        return min(hsp for hsp in self.hsps())

    def hasScoreBetterThan(self, score):
        """
        Is there an HSP with a score better than a given value?

        @return: A C{bool}, C{True} if there is at least one HSP in the
        alignments for this title with a score better than C{score}.
        """
        # Note: Do not assume that HSPs in an alignment are sorted in
        # decreasing order (as they are in BLAST output). If we could
        # assume that, we could just check the first HSP in each alignment.
        for hsp in self.hsps():
            if hsp.betterThan(score):
                return True
        return False

    def medianScore(self):
        """
        Find the median score for the HSPs in the alignments that match
        this title.

        @raise ValueError: If there are no HSPs.
        @return: The C{float} median score of HSPs in alignments matching the
            title.
        """
        return median([hsp.score.score for hsp in self.hsps()])

    def coverage(self):
        """
        Get the fraction of this title sequence that is matched by its reads.

        @return: The C{float} fraction of the title sequence matched by its
            reads.
        """
        intervals = ReadIntervals(self.subjectLength)
        for hsp in self.hsps():
            intervals.add(hsp.subjectStart, hsp.subjectEnd)
        return intervals.coverage()

    def residueCounts(self, convertCaseTo='upper'):
        """
        Count residue frequencies at all sequence locations matched by reads.

        @param convertCaseTo: A C{str}, 'upper', 'lower', or 'none'.
            If 'none', case will not be converted (both the upper and lower
            case string of a residue will be present in the result if they are
            present in the read - usually due to low complexity masking).
        @return: A C{dict} whose keys are C{int} offsets into the title
            sequence and whose values are C{Counters} with the residue as keys
            and the count of that residue at that location as values.
        """
        if convertCaseTo == 'none':
            convert = lambda x: x
        elif convertCaseTo == 'lower':
            convert = str.lower
        elif convertCaseTo == 'upper':
            convert = str.upper
        else:
            raise ValueError(
                "convertCaseTo must be one of 'none', 'lower', or 'upper'")

        counts = defaultdict(Counter)

        for titleAlignment in self:
            read = titleAlignment.read
            for hsp in titleAlignment.hsps:
                for (subjectOffset, residue, inMatch) in read.walkHSP(hsp):
                    counts[subjectOffset][convert(residue)] += 1

        return counts

    def summary(self):
        """
        Summarize the alignments for this subject.

        @return: A C{dict} with C{str} keys:
            bestScore: The C{float} best score of the matching reads.
            coverage: The C{float} fraction of the subject genome that is
                matched by at least one read.
            hspCount: The C{int} number of hsps that match the subject.
            medianScore: The C{float} median score of the matching reads.
            readCount: The C{int} number of reads that match the subject.
            subjectLength: The C{int} length of the subject.
            subjectTitle: The C{str} title of the subject.
        """
        return {
            'bestScore': self.bestHsp().score.score,
            'coverage': self.coverage(),
            'hspCount': self.hspCount(),
            'medianScore': self.medianScore(),
            'readCount': self.readCount(),
            'subjectLength': self.subjectLength,
            'subjectTitle': self.subjectTitle,
        }


class TitlesAlignments(dict):
    """
    Holds (as a dictionary) a set of titles, each with its alignments.

    @param readsAlignments: A L{dark.alignments.ReadsAlignments} instance.
    @param scoreClass: A class to hold and compare scores. If C{None},
        the score class from readsAlignments will be used.
    @param readSetFilter: An instance of dark.filter.ReadSetFilter, or C{None}.
        This can be used to pass a previously used title filter for ongoing
        use in filtering.
    @param importReadsAlignmentsTitles: If C{True}, titles from
        C{readsAlignments} will be added to self. This argument is only used
        by the filtering function to make a new instance without reading its
        titles.
    """

    def __init__(self, readsAlignments, scoreClass=None, readSetFilter=None,
                 importReadsAlignmentsTitles=True):
        dict.__init__(self)
        self.readsAlignments = readsAlignments
        self.scoreClass = scoreClass or readsAlignments.scoreClass
        self.readSetFilter = readSetFilter
        if importReadsAlignmentsTitles:
            for readAlignments in readsAlignments:
                for alignment in readAlignments:
                    title = alignment.subjectTitle
                    try:
                        titleAlignments = self[title]
                    except KeyError:
                        titleAlignments = self[title] = TitleAlignments(
                            title, alignment.subjectLength)
                    titleAlignments.addAlignment(
                        TitleAlignment(readAlignments.read, alignment.hsps))

    def addTitle(self, title, titleAlignments):
        """
        Add a new title to self.

        @param title: A C{str} title.
        @param titleAlignments: An instance of L{TitleAlignments}.
        @raises KeyError: If the title is already present.
        """
        if title in self:
            raise KeyError('Title %r already present in TitlesAlignments '
                           'instance.' % title)
        else:
            self[title] = titleAlignments

    def filter(self, minMatchingReads=None, minMedianScore=None,
               withScoreBetterThan=None, minNewReads=None, minCoverage=None,
               maxTitles=None, sortOn='maxScore'):
        """
        Filter the titles in self to create another TitlesAlignments.

        @param minMatchingReads: titles that are matched by fewer reads
            are unacceptable.
        @param minMedianScore: sequences that are matched with a median
            bit score that is less are unacceptable.
        @param withScoreBetterThan: if the best score for a title is not
            as good as this value, the title is not acceptable.
        @param minNewReads: The C{float} fraction of its reads by which a new
            title's read set must differ from the read sets of all previously
            seen titles in order for this title to be considered acceptably
            different (and therefore interesting).
        @param minCoverage: The C{float} minimum fraction of the title sequence
            that must be matched by at least one read.
        @param maxTitles: A non-negative C{int} maximum number of titles to
            keep. If more titles than this are present, titles will be sorted
            (according to C{sortOn}) and only the best will be retained.
        @param sortOn: A C{str} attribute to sort on, used only if C{maxTitles}
            is not C{None}. See the C{sortTitles} method below for the legal
            values.
        @raise: C{ValueError} if C{maxTitles} is less than zero or the value of
            C{sortOn} is unknown.
        @return: A new L{TitlesAlignments} instance containing only the
            matching titles.
        """
        # Use a ReadSetFilter only if we're checking that read sets are
        # sufficiently new.
        if minNewReads is None:
            readSetFilter = None
        else:
            if self.readSetFilter is None:
                self.readSetFilter = ReadSetFilter(minNewReads)
            readSetFilter = self.readSetFilter

        result = TitlesAlignments(
            self.readsAlignments, self.scoreClass, self.readSetFilter,
            importReadsAlignmentsTitles=False)

        if maxTitles is not None and len(self) > maxTitles:
            if maxTitles < 0:
                raise ValueError('maxTitles (%r) cannot be negative.' %
                                 maxTitles)
            else:
                # There are too many titles. Make a sorted list of them so
                # we loop through them (below) in the desired order and can
                # break when/if we've reached the maximum. We can't just
                # take the first maxTitles titles from the sorted list now,
                # as some of those titles might later be discarded by the
                # filter and then we'd return a result with fewer titles
                # than we should.
                titles = self.sortTitles(sortOn)
        else:
            titles = self.keys()

        for title in titles:
            # Test max titles up front, as it may be zero.
            if maxTitles is not None and len(result) == maxTitles:
                break

            titleAlignments = self[title]
            if (minMatchingReads is not None and
                    titleAlignments.readCount() < minMatchingReads):
                continue

            # To compare the median score with another score, we must
            # convert both values to instances of the score class used in
            # this data set so they can be compared without us needing to
            # know if numerically greater scores are considered better or
            # not.
            if (minMedianScore is not None and
                    self.scoreClass(titleAlignments.medianScore()) <
                    self.scoreClass(minMedianScore)):
                continue

            if (withScoreBetterThan is not None and not
                    titleAlignments.hasScoreBetterThan(withScoreBetterThan)):
                continue

            if (minCoverage is not None and
                    titleAlignments.coverage() < minCoverage):
                continue

            if (readSetFilter and not
                    readSetFilter.accept(title, titleAlignments)):
                continue

            result.addTitle(title, titleAlignments)

        return result

    def hsps(self):
        """
        Get all HSPs for all the alignments for all titles.

        @return: A generator yielding L{dark.hsp.HSP} instances.
        """
        return (hsp for titleAlignments in self.values()
                for alignment in titleAlignments for hsp in alignment.hsps)

    def sortTitles(self, by):
        """
        Sort titles by a given attribute and then by title.

        @param by: A C{str}, one of 'length', 'maxScore', 'medianScore',
            'readCount', or 'title'.
        @raise ValueError: If an unknown C{by} value is given.
        @return: A sorted C{list} of titles.
        """
        # First sort titles by the secondary key, which is always the title.
        titles = sorted(iter(self))

        # Then sort on the primary key (if any).
        if by == 'length':
            return sorted(
                titles,  reverse=True,
                key=lambda title: self[title].subjectLength)
        if by == 'maxScore':
            return sorted(
                titles, reverse=True, key=lambda title: self[title].bestHsp())
        if by == 'medianScore':
            return sorted(
                titles, reverse=True,
                key=lambda title: self.scoreClass(self[title].medianScore()))
        if by == 'readCount':
            return sorted(
                titles, reverse=True,
                key=lambda title: self[title].readCount())
        if by == 'title':
            return titles

        raise ValueError('Sort attribute must be one of "length", "maxScore", '
                         '"medianScore", "readCount", "title".')

    def summary(self, sortOn=None):
        """
        Summarize all the alignments for this title.

        @param sortOn: A C{str} attribute to sort titles on. One of 'length',
            'maxScore', 'medianScore', 'readCount', or 'title'.
        @raise ValueError: If an unknown C{sortOn} value is given.
        @return: A generator that yields C{dict} instances as produced by
            C{TitleAlignments} (see class earlier in this file), sorted by
            C{sortOn}.
        """
        titles = self if sortOn is None else self.sortTitles(sortOn)

        for title in titles:
            yield self[title].summary()

    def tabSeparatedSummary(self, sortOn=None):
        """
        Summarize all the alignments for this title as multi-line string with
        TAB-separated values on each line.

        @param sortOn: A C{str} attribute to sort titles on. One of 'length',
            'maxScore', 'medianScore', 'readCount', or 'title'.
        @raise ValueError: If an unknown C{sortOn} value is given.
        @return: A newline-separated C{str}, each line with a summary of a
            title. Each summary line is TAB-separated.
        """
        # The order of the fields returned here is somewhat arbitrary. The
        # subject titles are last because they are so variable in length.
        # Putting them last makes it more likely that the initial columns in
        # printed output will be easier to read down.
        #
        # Note that post-processing scripts will be relying on the field
        # ordering here.  So you can't just add fields. It's probably safe
        # to add them at the end, but be careful / think.
        #
        # A TAB-separated file can easily be read by awk using e.g.,
        # awk 'BEGIN {FS = "\t"} ...'

        result = []
        for titleSummary in self.summary(sortOn):
            result.append('\t'.join([
                '%(coverage)f',
                '%(medianScore)f',
                '%(bestScore)f',
                '%(readCount)d',
                '%(hspCount)d',
                '%(subjectLength)d',
                '%(subjectTitle)s',
            ]) % titleSummary)
        return '\n'.join(result)



import six
from unittest import TestCase
from dark.local_align import LocalAlignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


class TestLocalAlign(TestCase):
    """
    Test the LocalAlignment class.
    With match +1, mismatch -1, gap open -1, gap extend -1 and
        gap extend decay 0.0.
    """

    def testNonDNAString(self):
        """
        If the sequences do not consist of nucleotides, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq('xxx'), id='seq1')
        seq2 = SeqRecord(Seq('yyy'), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Invalid DNA nucleotide: "X"',
                              LocalAlignment, seq1, seq2)

    def testPositiveMismatch(self):
        """
        If the mismatch value passed is positive, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Mismatch must be negative',
                              LocalAlignment, seq1, seq2, mismatch=3)

    def testZeroMismatch(self):
        """
        If the mismatch value passed is zero, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Mismatch must be negative',
                              LocalAlignment, seq1, seq2, mismatch=0)

    def testPositiveGap(self):
        """
        If the gap value passed is positive, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Gap must be negative',
                              LocalAlignment, seq1, seq2, gap=3)

    def testZeroGap(self):
        """
        If the gap value passed is zero, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Gap must be negative',
                              LocalAlignment, seq1, seq2, gap=0)

    def testPositiveGapExtend(self):
        """
        If the gap extend value passed is positive, an exception
        must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError,
                              'Gap extension penalty cannot be positive',
                              LocalAlignment, seq1, seq2, gapExtend=3)

    def testFirstSequenceEmpty(self):
        """
        If the first sequence passed is empty, an exception must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq('agtcagtcagtc'), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq1',
                              LocalAlignment, seq1, seq2)

    def testSecondSequenceEmpty(self):
        """
        If the second sequence passed is empty, an exception must be raised.
        """
        seq1 = SeqRecord(Seq('agtcagtcagtc'), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq2',
                              LocalAlignment, seq1, seq2)

    def testBothSequencesEmpty(self):
        """
        If two empty sequences are passed, an exception must be raised.
        """
        seq1 = SeqRecord(Seq(''), id='seq1')
        seq2 = SeqRecord(Seq(''), id='seq2')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq1',
                              LocalAlignment, seq1, seq2)

    def testGapAtStartOfSeq1(self):
        seq1 = SeqRecord(Seq('gaatcg'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 6=\n"
                     "seq1 Match start: 1 Match end: 6\n"
                     "seq2 Match start: 2 Match end: 7\n"
                     "seq1 1 GAATCG 6\n"
                     "       ||||||\n"
                     "seq2 2 GAATCG 7")
        self.assertEqual(result, alignment)

    def testGapAtStartOfSeq2(self):
        seq1 = SeqRecord(Seq('cgaatcg'), id='seq1')
        seq2 = SeqRecord(Seq('gaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 6=\n"
                     "seq1 Match start: 2 Match end: 7\n"
                     "seq2 Match start: 1 Match end: 6\n"
                     "seq1 2 GAATCG 7\n"
                     "       ||||||\n"
                     "seq2 1 GAATCG 6")
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq1(self):
        seq1 = SeqRecord(Seq('cgaatc'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 6=\n"
                     "seq1 Match start: 1 Match end: 6\n"
                     "seq2 Match start: 1 Match end: 6\n"
                     "seq1 1 CGAATC 6\n"
                     "       ||||||\n"
                     "seq2 1 CGAATC 6")
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq2(self):
        seq1 = SeqRecord(Seq('cgaatcg'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatc'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 6=\n"
                     "seq1 Match start: 1 Match end: 6\n"
                     "seq2 Match start: 1 Match end: 6\n"
                     "seq1 1 CGAATC 6\n"
                     "       ||||||\n"
                     "seq2 1 CGAATC 6")
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq1(self):
        seq1 = SeqRecord(Seq('gaatc'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 5=\n"
                     "seq1 Match start: 1 Match end: 5\n"
                     "seq2 Match start: 2 Match end: 6\n"
                     "seq1 1 GAATC 5\n"
                     "       |||||\n"
                     "seq2 2 GAATC 6")
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq2(self):
        seq1 = SeqRecord(Seq('cgaatcg'), id='seq1')
        seq2 = SeqRecord(Seq('gaatc'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 5=\n"
                     "seq1 Match start: 2 Match end: 6\n"
                     "seq2 Match start: 1 Match end: 5\n"
                     "seq1 2 GAATC 6\n"
                     "       |||||\n"
                     "seq2 1 GAATC 5")
        self.assertEqual(result, alignment)

    def testAlignmentWithGapInMiddle(self):
        seq1 = SeqRecord(Seq('agtcagtcagtc'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 2=1D1=\n"
                     "seq1 Match start: 7 Match end: 10\n"
                     "seq2 Match start: 5 Match end: 7\n"
                     "seq1 7 TCAG 10\n"
                     "       || |\n"
                     "seq2 5 TC-G 7")
        self.assertEqual(result, alignment)

    def testTwoEqualSequences(self):
        """
        When two identical sequences are given, the result should
        show that the sequences completely match.
        """
        seq1 = SeqRecord(Seq('cgaatcg'), id='seq1')
        seq2 = SeqRecord(Seq('cgaatcg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 7=\n"
                     "seq1 Match start: 1 Match end: 7\n"
                     "seq2 Match start: 1 Match end: 7\n"
                     "seq1 1 CGAATCG 7\n"
                     "       |||||||\n"
                     "seq2 1 CGAATCG 7")
        self.assertEqual(result, alignment)

    def testTwoCompletelyDifferentSequences(self):
        """
        When two completely different sequences are given, the result
        should be the two sequences with an empty alignment.
        """
        seq1 = SeqRecord(Seq('aaaaaa'), id='seq1')
        seq2 = SeqRecord(Seq('gggggg'), id='seq2')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment()
        alignment = ("\nNo alignment between seq1 and seq2 \n")
        self.assertEqual(result, alignment)

    def testWikiAnswer(self):
        """
        Test the example given in Wikipedia:
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        """
        seq1 = SeqRecord(Seq('ACACACTA'), id='seq1')
        seq2 = SeqRecord(Seq('AGCACACA'), id='seq2')
        align = LocalAlignment(seq1, seq2, match=2)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 1=1I5=1D1=\n"
                     "seq1 Match start: 1 Match end: 8\n"
                     "seq2 Match start: 1 Match end: 8\n"
                     "seq1 1 A-CACACTA 8\n"
                     "       | ||||| |\n"
                     "seq2 1 AGCACAC-A 8")
        self.assertEqual(result, alignment)

    def testWikiAnswerWithMatchOne(self):
        """
        Test the example given in Wikipedia
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        Wikipedia uses a match score of two, here we use a score of one.
        """
        seq1 = SeqRecord(Seq('ACACACTA'), id='seq1')
        seq2 = SeqRecord(Seq('AGCACACA'), id='seq2')
        align = LocalAlignment(seq1, seq2, match=1)
        result = align.createAlignment()
        alignment = ("\nCigar string of aligned region: 5=1D1=\n"
                     "seq1 Match start: 2 Match end: 8\n"
                     "seq2 Match start: 3 Match end: 8\n"
                     "seq1 2 CACACTA 8\n"
                     "       ||||| |\n"
                     "seq2 3 CACAC-A 8")
        self.assertEqual(result, alignment)



from os.path import join, dirname

from cffi import FFI


def fileToString(filename):
    """
    Read a file's contents and return it as a string.

    @param filename: A C{str} file name.
    @return: A C{str} containing the contents of the file.
    """
    with open(join(dirname(__file__), filename)) as f:
        s = f.read()
    return s


ffi = FFI()

header = fileToString('gor4-base.h')

ffi.set_source('dark._gor4', header +
               fileToString('defines.h') +
               fileToString('nrutil.h') +
               fileToString('nrutil.c') +
               fileToString('gor4-base.c') +
               fileToString('api.c'))

ffi.cdef(header)

if __name__ == '__main__':
    ffi.compile()



#!/usr/bin/env python

"""
Read summary output produced by group-summary-proteins.py from many samples
and group it by virus.

Reads filenames from standard input, writes to standard output.

The filenames on stdin will typically be the 05-panel/summary-virus files.
E.g.,

    $ find . -name summary-virus | group-summary-viruses.py
"""

from __future__ import print_function

import sys
import re
from collections import defaultdict

SUMMARY_RE = re.compile(' (\(\d+ proteins?, \d+ reads?\))$')

viruses = defaultdict(list)
fileCount = 0

for filename in sys.stdin:
    filename = filename[:-1]
    fileCount += 1
    with open(filename) as fp:
        for line in fp:
            line = line[:-1]
            match = SUMMARY_RE.search(line)
            if match:
                virus = line[:-len(match.group(0))]
                counts = match.group(1)

                simpleFilename = filename
                if simpleFilename.startswith('./'):
                    simpleFilename = simpleFilename[2:]
                if simpleFilename.endswith('/05-panel/summary-virus'):
                    simpleFilename = simpleFilename[:-23]

                viruses[virus].append('  %s %s' % (simpleFilename, counts))
            elif line:
                viruses[virus].append('    %s' % (line,))

print('Proteins from %d viruses were found across %d samples.' %
      (len(viruses), fileCount))
print()

for virus in sorted(viruses):
    fileInfo = viruses[virus]
    sampleCount = len([f for f in fileInfo if not f.startswith('    ')])
    print('%s (in %d sample%s)' % (virus, sampleCount,
                                   '' if len(fileInfo) == 1 else 's'))
    print('\n'.join(fileInfo))
    print()



from time import time
from collections import defaultdict

# TODO: Some of these imported functions are no longer in utils.py!
from dark.utils import findHits, getSequence, summarizeHits, printHSP, report


def consensusSequence(recordFilename, hitId, fastaFilename, eCutoff=None,
                      db='nt', actualSequenceId=None):
    """
    Build a consensus sequence against a target sequence.

    recordFilename: the BLAST XML output file.
    hitId: the str sequence id to examine the BLAST output for hits against.
    fastaFilename: The name of the FASTA file containing query sequences,
        or a list of fasta sequences from SeqIO.parse
    eCutoff: converted e values less than this will be ignored.
    db: the BLAST db to use to look up the target and, if given, actual
        sequence.
    actualSequenceId: the str id of the actual sequence (if known).
    """

    print('TODO: This function is not finished yet.')
    return

    start = time()
    if isinstance(recordFilename, str):
        # TODO: REMOVE THE LIMIT IN THE NEXT LINE!
        allhits = findHits(recordFilename, set([hitId]), limit=100)
    else:
        allhits = recordFilename
    sequence = getSequence(hitId, db)
    fasta, summary = summarizeHits(allhits, fastaFilename, eCutoff=eCutoff)
    minX, maxX = summary['minX'], summary['maxX']
    if actualSequenceId:
        # UNUSED.
        # actualSequence = getSequence(actualSequenceId, db)
        pass
    print(summary['hitCount'])
    print('seq len =', len(sequence))
    fasta = summary['fasta']
    # The length of the consensus depends on where the query sequences fell
    # when aligned with the target. The consensus could extend the target
    # at both ends.
    consensusLen = maxX - minX
    consensus = [None, ] * consensusLen
    for item in summary['items']:
        print('NEW HSP')
        printHSP(item['origHsp'])  # TODO: REMOVE ME
        hsp = item['hsp']
        print('HIT query-start=%d query-stop=%d subj-start=%d subj-stop=%d' % (
            hsp['queryStart'], hsp['queryEnd'], hsp['subjectStart'],
            hsp['subjectEnd']))
        # print '   match: %s%s' % ('.' * hsp['subjectStart'], '-' *
        # (hsp['subjectEnd'] - hsp['subjectStart']))
        if item['frame']['subject'] > 0:
            query = fasta[item['sequenceId']].seq
        else:
            query = fasta[item['sequenceId']].reverse_complement().seq
        print('       target:',
              sequence[hsp['queryStart']:hsp['queryEnd']].seq)
        print('        query:', query)
        match = []
        for index in range(hsp['subjectStart'], hsp['subjectEnd']):
            queryIndex = index - hsp['queryStart']
            match.append('.' if query[queryIndex] == sequence[index] else '*')
        print('        match: %s%s' % (
            ' ' * (hsp['subjectStart'] - hsp['queryStart']), ''.join(match)))
        print('        score:', item['convertedE'])
        print('    match len:', hsp['subjectEnd'] - hsp['subjectStart'])
        print('subject frame:', item['frame']['subject'])
        for queryIndex, sequenceIndex in enumerate(
                range(hsp['queryStart'], hsp['queryEnd'])):
            consensusIndex = sequenceIndex + minX
            locus = consensus[consensusIndex]
            if locus is None:
                consensus[consensusIndex] = locus = defaultdict(int)
            locus[query[queryIndex]] += 1

    # Print the consensus before the target, if any.
    for index in range(minX, 0):
        consensusIndex = index - minX
        if consensus[consensusIndex]:
            print('%d: %r' % (index, consensus[consensusIndex]))
    # Print the consensus as it overlaps with the target, if any.
    for index in range(0, len(sequence)):
        consensusIndex = index - minX
        try:
            if consensus[consensusIndex]:
                print('%d: %r (%s)' % (
                    index, consensus[index], sequence[index]))
        except KeyError:
            # There's nothing left in the consensus, so we're done.
            break
    for index in range(len(sequence), maxX):
        consensusIndex = index - minX
        if consensus[consensusIndex]:
            print('%d: %r' % (index, consensus[consensusIndex]))
    stop = time()
    report('Consensus sequence generated in %.3f mins.' %
           ((stop - start) / 60.0))
    return summary, consensus



#!/usr/bin/env python

"""
Read DNA or AA FASTA from stdin and print AA ORF-only FASTA to stdout. I.e.,
the output FASTA will contain one sequence for each ORF found in each
sequence on stdin.

If a minimal ORF length is given, only print ORFs of at least that length
unless we are asked to allow ORFs that are open at at least one end.

Note that no start or stop codons will appear in the output. If you are
wanting to simply translate DNA FASTA to AA FASTA on a sequence-by-sequence
basis leaving in start/stop codons, use dna-to-aa.py instead.
"""

from __future__ import print_function

import sys
from os.path import basename
import argparse

from Bio.Data.CodonTable import TranslationError

from dark.fasta import FastaReads
from dark.reads import AARead, RNARead, DNARead

TYPE = {
    'aa': AARead,
    'dna': DNARead,
    'rna': RNARead,
}


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Convert DNA or AA FASTA to AA ORF-only FASTA',
        epilog='Given DNA or AA FASTA on stdin, output AA ORF-only FASTA to '
        'stdout. Optionally, filter by minimum required ORF length and allow '
        'the output of short ORFs that are open.'
    )

    parser.add_argument(
        '--allowOpenORFs', default=False, action='store_true',
        help='If True, ORFs that do not meet the length requirement '
        '(as specified by --minORFLength) will be output as long as '
        'they are open.')

    parser.add_argument(
        '--minORFLength', metavar='LEN', type=int, default=None,
        help='Only ORFs of at least this length will be written to stdout.')

    parser.add_argument(
        '--type', default='dna', choices=TYPE.keys(),
        help='The type of the bases in the stdin FASTA.')

    args = parser.parse_args()
    write = sys.stdout.write
    minORFLength = args.minORFLength
    allowOpenORFs = args.allowOpenORFs

    reads = FastaReads(sys.stdin, readClass=TYPE[args.type])

    for read in reads:
        if args.type == 'aa':
            # No need to translate to AA, since our input is already AA.
            translations = [read]
        else:
            translations = read.translations()
        try:
            for translation in translations:
                for orf in translation.ORFs():
                    if minORFLength is None or len(orf) >= minORFLength or (
                            allowOpenORFs and (orf.openLeft or orf.openRight)):
                        write(orf.toString('fasta'))
        except TranslationError as error:
            print('Could not translate read %r sequence %r (%s).\nDid you '
                  'forget to run %s with "--type aa"?' %
                  (read.id, read.sequence, error, basename(sys.argv[0])),
                  file=sys.stderr)
            sys.exit(1)
